const int chA1 = 4;               // ch.A of the encoder measuring angle
const int chB1 = 2;               // ch.B of the encoder measuring angle
const int chA2 = 5;               // ch.A of the encoder measuring position
const int chB2 = 3;               // ch.B of the encoder measuring position
volatile int   chA1_value = 0;    // initial value of angle
volatile float angle_count = 0;   // counted number for angle
volatile float angle = 0;         // unit: radian
volatile int   chA2_value = 0;    // initial value of position
volatile float pos_count  = 0;    // counted number for position
volatile float pos = 0;           // position, unit: 10^(-4) meter
unsigned int   tcnt2;             // reload value of Timer2
int t = 0;                        // used in Timer2
unsigned int i = 0;               // used in DAC and moving average
unsigned char b[8] = {0,0,0,0,0,0,0,0}; // stores binary value
float u = 0;                      // input voltage
int digital_input = 0;            // digital input, will be converted to analog
float k1, k2, k3, k4;             // system feedback gains
float mid = 5.1;                  // voltage for valve that make no movement
float offset = 0.6;               // offset from "mid", essentially saturation limit
volatile float angle1 = 0;        // angle previous value
volatile float pos1 = 0;          // posotion previous value
volatile float angular_velocity = 0; // estimated angular velocity
volatile float cart_velocity = 0; // estimated cart velocity
const int samples = 5;            // number of samples for moving average
volatile float angRunAvg[samples] = {0,0,0,0,0}; //angular moving average values
volatile float posRunAvg[samples] = {0,0,0,0,0}; //cart velocity moving average values
volatile int ind = 0;             //index for moving average
volatile int minSamplePeriod = 4; //sampling period

void setup(){
  //Serial.begin(9600); 
  pinMode(chA1, INPUT); pinMode(chA2, INPUT); // for encoders
  attachInterrupt(0, check_chA1, RISING);     // for angle encoder
  attachInterrupt(1, check_chA2, FALLING);    // for position encoder
  
  // for DAC
  pinMode(A0, OUTPUT);pinMode(A1, OUTPUT);
  pinMode(A2, OUTPUT);pinMode(A3, OUTPUT);
  pinMode(A4, OUTPUT);pinMode(A5, OUTPUT);
  pinMode(6, OUTPUT); pinMode(7, OUTPUT);
  
  // Timer2 setup: overflows every 1 ms
  TIMSK2 &= ~(1<<TOIE2);
  TCCR2A &= ~((1<<WGM21) | (1<<WGM20));
  TCCR2B &= ~(1<<WGM22);
  ASSR &= ~(1<<AS2);
  TIMSK2 &= ~(1<<OCIE2A);
  TCCR2B |= (1<<CS22)  | (1<<CS20);
  TCCR2B &= ~(1<<CS21);
  tcnt2 = 131;
  TCNT2 = tcnt2;
  TIMSK2 |= (1<<TOIE2);
  
  // system feedback gains (results from Matlab)
  k1 = -54.01;
  k2 = -0.5043;
  k3 = 1.414;
  k4 = 11.94;
}

void loop(){
  if (angle < -0.007 || angle > 0.007){
    u = mid-(k1*angle+k2*angular_velocity+k3*pos+k4*cart_velocity);
    if (u<(mid-offset)){
      u = mid-offset;
    }
    else if (u>(mid+offset)){
      u = mid+offset;
    }
  }
  else {
    u = mid;
  }
  digital_input = (int) (256*u/12);
  /*
  Serial.print(angle, 4);
  Serial.print(" ");
  Serial.print(angular_velocity, 4);
  Serial.print(" ");
  Serial.print(pos, 4);
  Serial.print(" ");
  Serial.print(cart_velocity, 4);
  Serial.println(" ");
  */
  dec2bin(digital_input);
}
